import os
import time
import re
from datetime import datetime
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from sklearn.svm import SVR
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, explained_variance_score
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import KFold
from common import logger
from common import containers_info

# Set environment variables and initialize logging
global dataset
dataset = os.environ.get("DATASET")
logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)

# Initialize data variables and models
x_train, y_train, x_test, y_test, scaler = [], [], [], [], None
occupied = 0
global_filename, global_date = "", ""

modelSVM = SVR(kernel='rbf', C=1.0, epsilon=0.1)

def get_parameters():
    log.info("Getting Parameters!")
    return modelSVM.get_weights()

def set_parameters(parameters):
    modelSVM.set_weights(parameters)
    log.info("Parameters Set!")

def in_usage(value):
    global occupied
    occupied = value
    log.info("ML app is now in use" if value == 1 else "ML app is now freed")

def split_data(df, split_ratio=0.75):
    global x_train, y_train, x_test, y_test
    split_index = int(len(df) * split_ratio)
    x_train = df.iloc[:split_index][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_train = df.iloc[:split_index]["car_sum_scaler"]
    x_test = df.iloc[split_index:][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_test = df.iloc[split_index:]["car_sum_scaler"]

def get_data():
    ##########################################################################################
    # ASK TO ANOTHER ENTITY WHICH IS THE DATASET FILE AND OTHER RESTRICTIONS
    file_name = dataset
    date = "2022-05-10"
    ##########################################################################################

    global global_filename
    global_filename = str(file_name).split(".")[0]
    global global_date
    global_date = str(date)
    global scaler
    df = pd.read_csv("/app/dataset/" + dataset, index_col="timestamp", parse_dates=True)
    scaler = MinMaxScaler()
    df["car_sum_scaler"] = scaler.fit_transform(df["car_sum"].to_numpy().reshape(-1, 1))
    split_index = int(len(df) * 0.25)
    df["car_sumB_scaler"] = df["car_sum_scaler"].shift(split_index)
    df["car_sumBB_scaler"] = df["car_sum_scaler"].shift(split_index * 2)
    df["car_sumBBB_scaler"] = df["car_sum_scaler"].shift(split_index * 3)
    df.fillna(0, inplace=True)
    split_data(df)
    return len(x_train)

def fit(config):
    kf = KFold(n_splits=5, shuffle=True, random_state=0)
    fold_results = []

    for train_index, val_index in kf.split(x_train):
        x_train_fold, x_val_fold = x_train.iloc[train_index], x_train.iloc[val_index]
        y_train_fold, y_val_fold = y_train.iloc[train_index], y_train.iloc[val_index]
        modelSVM.fit(x_train_fold, y_train_fold)
        val_predictions = modelSVM.predict(x_val_fold)
        fold_results.append(mean_squared_error(y_val_fold, val_predictions))

    avg_mse = np.mean(fold_results)
    return len(x_train), {"avg_mse": avg_mse}

def evaluate():
    predictions = modelSVM.predict(x_test)
    test_mse = mean_squared_error(y_test, predictions)
    test_rmse = np.sqrt(test_mse)
    test_mae = mean_absolute_error(y_test, predictions)
    test_r2 = r2_score(y_test, predictions)

    return {
        "test_mse": test_mse,
        "test_rmse": test_rmse,
        "test_mae": test_mae,
        "test_r2_score": test_r2,
    }

def predict(before_after, plot_graphs=False):
    predictions = modelSVM.predict(x_test)
    result = scaler.inverse_transform(predictions.reshape(-1, 1))
    real_values = scaler.inverse_transform(y_test.values.reshape(-1, 1))

    test_mse = mean_squared_error(real_values, result)
    test_rmse = np.sqrt(test_mse)
    test_mae =  mean_absolute_error(real_values, result)
    test_r2_score = r2_score(real_values, result)
    test_explained_variance = explained_variance_score(real_values, result)

    # Create today's path based on the current date
    today = datetime.now()
    today_path = f"/app/logs/SVM_{dataset}_{today}"
    if not os.path.exists(today_path):
        os.makedirs(today_path)

    # Define the log file path
    log_file = os.path.join(today_path, 'metrics_log.txt')

    # Write metrics to the file
    with open(log_file, 'a') as f:
        f.write(f"Model: SVM\n")
        f.write(f"Dataset: {global_filename}\n")
        f.write(f"Evaluation MSE: {test_mse}\n")
        f.write(f"Evaluation RMSE: {test_rmse}\n")
        f.write(f"Evaluation MAE: {test_mae}\n")
        f.write(f"Evaluation R2SCORE: {test_r2_score}\n")
        f.write(f"Evaluation Explained Variance: {test_explained_variance}\n")
        f.write("\n")  # Add a newline for separation between log entries
    
    results ={
        "loss": 0,
        "r2_score":0,
        "val_loss":0,
        "val_r2_score": 0,
    }
     # Plot predictions vs actual values
    ARIMAplot_graphs_fn(str(global_filename), str(global_date), real_values, result, today_path, today)

    return test_r2_score, results

def ARIMAplot_graphs_fn(file_name, global_date, ARIMA_history, predictions, today_path, today):
    log.info(str(ARIMA_history))
    log.info(str(predictions))
    plt.figure(figsize=(20, 10))
    plt.plot(ARIMA_history, label="real")
    plt.plot(predictions, label="prediction")
    plt.legend()
    plt.title("SVM prediction")
    
    predictions['car_sum'].to_csv(f"{today_path}/{today}.csv")

    # Regular expression to match files like "run0.json", "run1.json", etc.
    pattern = re.compile(r'^run(\d+)\.json$')

    # Find all files that match the pattern and extract the number part
    count = -1
    for path in os.listdir(today_path):
        if os.path.isfile(os.path.join(today_path, path)):
            if pattern.match(path):
                count += 1

    count = 0 if count == -1 else count

    runi = f"run{count}"

    plt.savefig(today_path + "/predict_" + file_name + "_" + runi + ".png")


def get_prediction_for_day_hour(matrix, day, hour):
    index = day * 24 + hour
    if index < 0 or index >= matrix.shape[0]:
        raise ValueError("Invalid day or hour value.")
    return matrix[index, 0]

def predictOne(day, hour):
    result_scaler = modelSVM.predict(x_test)
    result = scaler.inverse_transform(result_scaler)
    return {"prediction": str(get_prediction_for_day_hour(result, day, hour))}

def evaluate(config):
    log.info("----------------  EVALUATE  ----------------- ")
    log.debug("config: " + str(config))

    global modelA
    global x_test
    global y_test

    start_time_evaluate = time.time()
    loss, r_square = modelA.evaluate(x_test, y_test, steps=int(config["val_steps"]))
    log.info("Evaluate Duration: " + str((time.time() - start_time_evaluate)))

    return float(loss), len(x_test), {"r2_score": float(r_square)}
