import json
import time
import os
import requests
from urllib.parse import quote

from fastapi import FastAPI, Query
from pydantic import BaseModel

from common import logger
from common import encoder_decoder
from common import containers_info

from clientML import evaluate as evaluate_M
from clientML import fit as fit_M
from clientML import get_data as get_data_M
from clientML import in_usage as in_usage_M
from clientML import predict as predict_M
from clientML import predictOne as predictOne_M

logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))

# [] -> actual parameters, 0 -> value of the metric
parameters_best = ([], 0)
parameters_before_last_agg = ([], 0)

# Configure logging
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)
logs = log.log

mflinterface_app = FastAPI()

class Body(BaseModel):
    parameters: str = None

@mflinterface_app.get("/get_data")
def get_data(client_n: str = Query(description="e.g. 1", default="1")):
    # GET TIMES
    tc2 = time.time()
    logs(5, "TC2_"+ str(client_n) +"=" + str(tc2))

    global parameters_best
    global parameters_before_last_agg

    len_x_train = get_data_M()

    # GET TIMES
    tc3 = time.time()
    logs(5, "TC3_"+ str(client_n) +"=" + str(tc3))
    log.info("DATARETRIEVED")
    return {"Stats": "Get Data done successfully!"}


@mflinterface_app.post("/fit")
def fit(
    body: Body,
    config: str = Query(
        description="config for fit function (string url safe)",
        default="%7B%22batch_size%22%3A%20%2232%22%2C%20%22epoch_global%22%3A%20%221%22%2C%20%22epochs%22%3A%20%2230%22%7D",
    ),
    current_or_server: str = Query(
        description="Use current parameters of the model or the parameters sent by the server",
        default="0"
    ),
    client_n: str = Query(description="e.g. 1", default="1"),
):
    # GET TIMES
    tr3 = time.time()

    # Parsing the url parameters to their real data structure
    # Convert the config string back to a dictionary
    config_json = encoder_decoder.decode_config(config_url=config)
    logs(5, "T_R"+ str(config_json["epoch_global"]) +"_3_"+ str(client_n) +"=" + str(tr3))

    # TRAIN THE MODEL
    num_examples, metrics = fit_M(config=config_json)

    # Convert results dictionary to a url safe string
    metrics_string = json.dumps(metrics)
    metrics_encoded_string = quote(metrics_string)

    # GET TIMES
    tr4 = time.time()
    logs(5, "T_R"+ str(config_json["epoch_global"]) +"_4_"+ str(client_n) +"=" + str(tr4))

    return {
        "num_examples": num_examples,
        "metrics": str(metrics_encoded_string),
    }


@mflinterface_app.post("/evaluate")
def evaluate(
    body: Body,
    config: str = Query(
        description="config for evaluate function (string url safe)", default="%7B%22val_steps%22%3A%225%22%2C%22epoch_global%22%3A%221%22%7D"
    ),
    current_or_server: str = Query(
        description="Use current parameters of the model or the parameters sent by the server",
        default="0"
    ),
    client_n: str = Query(description="e.g. 1", default="1"),
):
    # GET TIMES
    tr11 = time.time()

    # Parsing the url parameters to their real data structure
    # Convert the config string back to a dictionary
    config_json = encoder_decoder.decode_config(config_url=config)
    logs(5, "T_R"+ str(config_json["epoch_global"]) +"_11_"+ str(client_n) +"=" + str(tr11))
    
  
    # EVALUATE THE MODEL
    losses, x_test_len, metrics = evaluate_M(config=config_json)

    # Convert results dictionary to a url safe string
    metrics_string = json.dumps(metrics)
    metrics_encoded_string = quote(metrics_string)

    # GET TIMES
    tr12 = time.time()
    logs(5, "T_R"+ str(config_json["epoch_global"]) +"_12_"+ str(client_n) +"=" + str(tr12))

    return {
        "losses": float(losses),
        "num_examples": x_test_len,
        "metrics": str(metrics_encoded_string),
    }


@mflinterface_app.get("/predict/local")
def predict_local_model(plot_graphs: str = Query(description="true or false", default="false")):
    # TODO: FALTA MEDIR A ACCURACY


    # PREDICT THE MODEL WITH PREVIOUS PARAMETERS
    losses, metrics = predict_M(plot_graphs=str(plot_graphs), before_after="before")


    return {"losses": str(losses), "metrics": str(metrics), "container_id": containers_info.get_current_container_id()}


@mflinterface_app.get("/predictOne")
def predict_local_model(plot_graphs: str = Query(description="true or false", default="false"),
    day: str = Query(description="day of the week", default="2"),hour: str = Query(description="hour of the week", default="2")):
    # TODO: FALTA MEDIR A ACCURACY


    # PREDICT THE MODEL WITH PREVIOUS PARAMETERS
    predictions = predictOne_M(plot_graphs=str(plot_graphs), before_after="before")


    return {"losses": str(predictions),"container_id": containers_info.get_current_container_id()}



@mflinterface_app.get("/predict/aggregated")
def predict_aggregated_model(plot_graphs: str = Query(description="true or false", default="false")):
    # TODO: FALTA MEDIR A ACCURACY

    # PREDICT THE MODEL WITH CURRENT/NEW PARAMETERS
    losses, metrics = predict_M(plot_graphs=str(plot_graphs), before_after="after")

    return {"losses": str(losses), "metrics": str(metrics), "container_id": containers_info.get_current_container_id()}


@mflinterface_app.post("/free_ml")
def free_ml():
    # GET TIMES
    t16 = time.time()
    logs(5, "T16=" + str(t16))

    in_usage_M(0)

    # GET TIMES
    t17 = time.time()
    logs(5, "T17=" + str(t17))

    return {"Stats": "Aggregation ended, ML app free."}


@mflinterface_app.post("/start_aggregation")
def start_aggregation(
    n_rounds: str = Query(description="e.g. 5", default="5"),
    round_timeout: str = Query(description="e.g. 20.0 (if 0, there will be no timeout)", default="20.0"),
    fl_algorithm: str = Query(description="e.g. FedAvg", default="FedAvg"),
    n_epochs: str = Query(description="e.g. 30", default="30"),
    batch_size: str = Query(description="e.g. 32", default="32"),
    neighbours_file: str = Query(description="e.g. test1.json", default="test1.json"),
):
    # GET TIMES
    t0 = time.time()
    logs(5, "T0=" + str(t0))

    in_usage_M(1)

    # Get ml-app container name
    local_ml_app_container_name = containers_info.get_current_container_name()

    # Manager container name
    manager_container_name = "mfl-manager-" + local_ml_app_container_name

    params = {
        "ml_app_address": str(str(local_ml_app_container_name) + ":" + str(containers_info.get_current_container_internal_port()[0])),
        "manager_external_port": str(containers_info.get_exposed_port_named_container(manager_container_name)[0]),
        "n_rounds": str(n_rounds),
        "round_timeout": str(round_timeout),
        "fl_algorithm": str(fl_algorithm),
        "n_epochs": str(n_epochs),
        "batch_size": str(batch_size),
        "neighbours_file": str(neighbours_file),
    }
    url = "http://" + manager_container_name + ":" + containers_info.get_internal_port_named_container(manager_container_name)[0] + "/trigger_start_aggregation"

    try:
        # GET TIMES
        t1 = time.time()
        logs(5, "T1=" + str(t1))
        response = requests.post(url, params=params)

        # Proccess response
        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # GET TIMES
            t8 = time.time()
            logs(5, "T8=" + str(t8))

            log.info(url + " (SUCCESSFUL)")

        else:
            log.error(url + " (UNSUCCESSFUL): " + str(response.status_code))
    except requests.exceptions.RequestException as e:
        log.error(e)

    # GET TIMES
    t9 = time.time()
    logs(5, "T9=" + str(t9))

    return {"Stats": "Start Aggregation triggered on mfl-manager-" + local_ml_app_container_name + " !"}



